@@ -319,13 +319,34 @@ initialize_univariate_channel <- function(dformula, specials, fixed_pars,
   )
   channel <- c(channel, indices)
   sampling <- setNames(indices, paste0(names(indices), "_", y_name))
+  # evaluate specials such as offset and trials
+  spec_na <- matrix(FALSE, nrow = T_full - fixed, ncol = N)
+  for (spec in formula_special_funs) {
+    if (!is.null(specials[[spec]])) {
+      spec_idx <- seq.int(fixed + 1L, T_full)
+      spec_split <- split(specials[[spec]], group)
+      spec_array <- array(as.numeric(unlist(spec_split)), dim = c(T_full, N))
+      spec_na <- spec_na | is.na(spec_array[spec_idx, , drop = FALSE])
+      spec_name <- paste0(spec, "_", y_name)
+      sampling[[spec_name]] <- ifelse_(
+        identical(spec, "offset"),
+        t(spec_array[spec_idx, , drop = FALSE]),
+        spec_array[spec_idx, , drop = FALSE]
+      )
+      channel[[paste0("has_", spec)]] <- TRUE
+    } else {
+      channel[[paste0("has_", spec)]] <- FALSE
+    }
+  }
   obs_idx <- array(0L, dim = c(N, T_full - fixed))
   obs_len <- integer(T_full - fixed)
   for (j in seq_len(T_full - fixed)) {
     x_na <- X_na[j, , channel$J, drop = FALSE]
     dim(x_na) <- c(N, channel$K)
     y_na <- Y_na[j, ]
-    obs_XY <- which(apply(x_na, 1L, function(z) all(!z)) & !y_na)
+    obs_XY <- which(
+      apply(x_na, 1L, function(z) all(!z)) & !y_na & all(!spec_na[j, ])
+    )
     obs_XY_len <- length(obs_XY)
     obs_idx[, j] <- c(obs_XY, rep(0L, N - obs_XY_len))
     obs_len[j] <- obs_XY_len
@@ -336,11 +357,6 @@ initialize_univariate_channel <- function(dformula, specials, fixed_pars,
   sampling[[paste0("n_obs_", y_name)]] <- obs_len
   sampling[[paste0("t_obs_", y_name)]] <- which(obs_len > 0L)
   sampling[[paste0("T_obs_", y_name)]] <- length(which(obs_len > 0L))
-  channel$t_obs <- ifelse_(
-    channel$has_fully_missing,
-    glue::glue("t_obs_{y_name}"),
-    "1:T"
-  )
   # obs selects complete cases if there are missing observations
   channel$obs <- ifelse_(
     channel$has_missing,
@@ -366,18 +382,6 @@ initialize_univariate_channel <- function(dformula, specials, fixed_pars,
     "Model for response variable {.var {y}} contains time-varying
      definitions but splines have not been defined."
   )
-  # evaluate specials such as offset and trials
-  for (spec in formula_special_funs) {
-    if (!is.null(specials[[spec]])) {
-      spec_split <- split(specials[[spec]], group)
-      spec_array <- array(as.numeric(unlist(spec_split)), dim = c(T_full, N))
-      sampling[[paste0(spec, "_", y_name)]] <-
-        spec_array[seq.int(fixed + 1L, T_full), , drop = FALSE]
-      channel[[paste0("has_", spec)]] <- TRUE
-    } else {
-      channel[[paste0("has_", spec)]] <- FALSE
-    }
-  }
   sampling[[paste0("y_", y_name)]] <- ifelse_(
     dformula$family %in%
       c("gaussian", "gamma", "exponential", "beta", "student"),
@@ -414,11 +418,6 @@ initialize_multivariate_channel <- function(y, y_cg, y_name, cg_idx,
   )
   channel$y <- unname(vapply(channel_vars[cg_idx], "[[", character(1L), "y"))
   channel$y_cg <- y_cg
-  channel$t_obs <- ifelse_(
-    channel$has_fully_missing,
-    glue::glue("t_obs_{y_cg}"),
-    "1:T"
-  )
   channel$obs <- ifelse_(
     channel$has_missing,
     glue::glue("obs_{y_cg}[1:n_obs_{y_cg}[t], t]"),
@@ -431,6 +430,8 @@ initialize_multivariate_channel <- function(y, y_cg, y_name, cg_idx,
       channel[[has_spec]] <- TRUE
       sampling[[paste0(spec, "_", y_cg)]] <-
         sampling_vars[[paste0(spec, "_", z)]]
+    } else {
+      channel[[has_spec]] <- FALSE
     }
   }
   sampling[[paste0("obs_", y_cg)]] <- matrix_intersect(
