@@ -12,15 +12,16 @@
 #' @param mvars \[`list()`]\cr The `model_vars` component of
 #'   [prepare_stan_input()] output.
 #' @noRd
-create_blocks <- function(indent = 2L, backend, cg, cvars, cgvars, mvars) {
+create_blocks <- function(indent = 2L, backend, cg, cvars, cgvars, mvars,
+                          threading) {
   idt <- indenter_(indent)
   paste_rows(
-    create_functions(idt, backend, cg, cvars, cgvars, mvars),
-    create_data(idt, backend, cg, cvars, cgvars, mvars),
+    create_functions(idt, backend, cg, cvars, cgvars, mvars, threading),
+    create_data(idt, backend, cg, cvars, cgvars, mvars, threading),
     create_transformed_data(idt, backend, cg, cvars, cgvars, mvars),
     create_parameters(idt, backend, cg, cvars, cgvars, mvars),
     create_transformed_parameters(idt, backend, cg, cvars, cgvars, mvars),
-    create_model(idt, backend, cg, cvars, cgvars, mvars),
+    create_model(idt, backend, cg, cvars, cgvars, mvars, threading),
     create_generated_quantities(idt, backend, cg, cvars, cgvars, mvars),
     .parse = FALSE
   )
@@ -32,7 +33,7 @@ create_blocks <- function(indent = 2L, backend, cg, cvars, cgvars, mvars) {
 #' @param idt \[`function`]\cr
 #'   An indentation function created by [indenter_()].
 #' @noRd
-create_functions <- function(idt, backend, cg, cvars, cgvars, mvars) {
+create_functions <- function(idt, backend, cg, cvars, cgvars, mvars, threading) {
   functions_psi <- ""
   psis <- mvars$lfactor_def$responses
   P <- mvars$lfactor_def$P
@@ -61,17 +62,49 @@ create_functions <- function(idt, backend, cg, cvars, cgvars, mvars) {
       .indent = idt(c(1, 2, 2, 3, 3, 2, 2, 1, 1, 2, 2, 2, 2, 3, 3, 2, 2, 2, 1))
     )
   }
+  likelihood_functions_text <- ""
+  if (threading) {
+    n_cg <- n_unique(cg)
+    likelihood_functions_text <- character(n_cg)
+    for (i in seq_len(n_cg)) {
+      cg_idx <- which(cg == i)
+      likelihood_functions_text[i] <- create_functions_lines(
+        idt, backend, cvars[cg_idx], cgvars[[i]], threading
+      )
+    }
+  }
   paste_rows(
     "functions {",
     functions_psi,
+    likelihood_functions_text,
     "}",
     .parse = FALSE
   )
 }
 
+#' Create Functions Lines for a Distribution
+#'
+#' @noRd
+create_functions_lines <- function(idt, backend, cvars, cgvars, threading) {
+  family <- cgvars$family
+  if (is_multivariate(family)) {
+    lines_wrap("functions", family, idt, backend,
+               list(cvars = cvars, cgvars = cgvars, threading = threading))
+  } else {
+    if (is_categorical(family)) {
+      cvars[[1L]]$threading <- threading
+    } else {
+      cvars[[1L]]$default <- lines_wrap(
+        "loglik", "default", idt, backend,
+        c(cvars[[1L]], threading = threading)
+      )
+    }
+    lines_wrap("functions", family, idt, backend, cvars[[1L]])
+  }
+}
 #' @describeIn create_function Create The 'Data' Block of the Stan Model Code
 #' @noRd
-create_data <- function(idt, backend, cg, cvars, cgvars, mvars) {
+create_data <- function(idt, backend, cg, cvars, cgvars, mvars, threading) {
   has_splines <- any(vapply(cvars, "[[", logical(1L), "has_varying")) ||
     any(vapply(cvars, "[[", logical(1L), "has_varying_intercept")) ||
     any(vapply(cvars, "[[", logical(1L), "has_lfactor"))
@@ -110,6 +143,7 @@ create_data <- function(idt, backend, cg, cvars, cgvars, mvars) {
       P > 0L,
       "int<lower=0> P; // number of channels with latent factor"
     ),
+    onlyif(threading, "int<lower=1> grainsize;"),
     .indent = idt(1),
     .parse = FALSE
   )
@@ -171,9 +205,14 @@ create_transformed_data <- function(idt, backend, cg, cvars, cgvars, mvars) {
     "transformed data {",
     declarations,
     onlyif(has_lfactor, "vector[2 * N] QR_Q = create_Q(N);"),
+    ifelse_(
+      stan_supports_array_keyword(backend),
+      "array[T] int seq1T = linspaced_int_array(T, 1, T);",
+      "int seq1T[T] = linspaced_int_array(T, 1, T);"
+    ),
     statements,
     "}",
-    .indent = idt(c(0, 0, 1, 0, 0)),
+    .indent = idt(c(0, 0, 1, 1, 0, 0)),
     .parse = FALSE
   )
 }
@@ -190,7 +229,7 @@ create_transformed_data_lines <- function(idt, backend, cvars, cgvars) {
         lines_wrap("transformed_data", "default", idt, backend, x)
       }
     )
-    tr_data <- lines_wrap("transformed_data", family, idt, backend, cgvars)
+    lines_wrap("transformed_data", family, idt, backend, cgvars)
   } else {
     cvars[[1L]]$default <- lines_wrap(
       "transformed_data", "default", idt, backend, cvars[[1L]]
@@ -477,13 +516,13 @@ create_transformed_parameters_lines <- function(idt, backend, cvars, cgvars) {
 
 #' @describeIn create_function Create the 'Model' Block of the Stan Model Code
 #' @noRd
-create_model <- function(idt, backend, cg, cvars, cgvars, mvars) {
+create_model <- function(idt, backend, cg, cvars, cgvars, mvars, threading) {
   spline_def <- mvars$spline_def
   spline_text <- ""
   if (!is.null(spline_def) && spline_def$shrinkage) {
     xi_prior <- mvars$common_priors
     xi_prior <- xi_prior[xi_prior$parameter == "xi", "prior"]
-    spline_text <- paste_rows("xi[1] ~ {xi_prior};", .indent = idt(1))
+    spline_text <- paste_rows("xi ~ {xi_prior};", .indent = idt(1))
   }
   random_text <- ""
   if (mvars$random_def$M > 0L) {
@@ -629,7 +668,7 @@ create_model <- function(idt, backend, cg, cvars, cgvars, mvars) {
   for (i in seq_len(n_cg)) {
     cg_idx <- which(cg == i)
     model_text[i] <- create_model_lines(
-      idt, backend, cvars[cg_idx], cgvars[[i]]
+      idt, backend, cvars[cg_idx], cgvars[[i]], threading = threading
     )
   }
   paste_rows(
@@ -646,7 +685,7 @@ create_model <- function(idt, backend, cg, cvars, cgvars, mvars) {
 #' Create Model Lines for a Distribution
 #'
 #' @noRd
-create_model_lines <- function(idt, backend, cvars, cgvars, mvars) {
+create_model_lines <- function(idt, backend, cvars, cgvars, mvars, threading) {
   family <- cgvars$family
   if (is_multivariate(family)) {
     cgvars$backend <- backend
@@ -655,7 +694,7 @@ create_model_lines <- function(idt, backend, cvars, cgvars, mvars) {
       family,
       idt,
       backend,
-      list(cvars = cvars, cgvars = cgvars)
+      list(cvars = cvars, cgvars = cgvars, threading = threading)
     )
   } else {
     cvars[[1L]]$backend <- backend
@@ -669,18 +708,13 @@ create_model_lines <- function(idt, backend, cvars, cgvars, mvars) {
         }
       )
       cvars[[1L]]$backend <- backend
-      cvars[[1L]]$intercept <- lapply(
-        cvars[[1L]]$categories[-1L],
-        function(s) {
-          cvars[[1L]]$ydim <- cvars[[1L]]$y
-          cvars[[1L]]$y <- paste0(cvars[[1L]]$y, "_", s)
-          do.call(intercept_lines, cvars[[1L]])
-        }
-      )
     } else {
       cvars[[1L]]$priors <- do.call(prior_lines, c(cvars[[1L]], idt = idt))
-      cvars[[1L]]$intercept <- do.call(intercept_lines, cvars[[1L]])
+      cvars[[1L]]$default <- lines_wrap(
+        "loglik", "default", idt, backend, c(cvars[[1L]], threading = threading)
+      )
     }
+    cvars[[1L]]$threading <- threading
     lines_wrap("model", family, idt, backend, cvars[[1L]])
   }
 }
